/*
    MIDICodes.h : Generic MIDI Codes
    Copyright (c) 2004-2016 Adrian M. Gin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License v2 as published by
    the Free Software Foundation;

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

*/


#ifndef MIDI_CODES_H
#define MIDI_CODES_H


#define OCTAVE_OFFSET      (1)
#define MIDI_SYSTEM_COMMON_MSG (0xF0)
#define MIDI_MSG_TYPE_MASK	(0xF0)
#define MIDI_CHANNEL_MASK	(0x0F)
#define MIDI_DATA_READY     (0x01)
#define MIDI_DATA_NOT_READY (0x00)

#define UNKNOWN_SYS_MSG (0xFF)
#define UNKNOWN_VOICE_MSG (0xFE)

/* In 100us resolution */
#define MIDI_MAX_CHANNELS (16)
#define MIDI_MAX_DATA   (127)
#define MIDI_MAX_KEY    (127)

#define MIDI_DRUM_CHANNEL   (0x09)

//For a Piano
#define MIDI_PIANO_A0 (21)
#define MIDI_PIANO_C8 (MIDI_PIANO_A0+MIDI_PIANO_KEY_COUNT-1)
#define MIDI_PIANO_KEY_COUNT (88)


typedef enum {
   MIDI_NOTE_OFF = 0x80,
   MIDI_NOTE_ON = 0x90,
   MIDI_AFTERTOUCH = 0xA0,
   MIDI_CONTROL_CHANGE = 0xB0,
   MIDI_PROGRAM_CHANGE = 0xC0,
   MIDI_CHANNEL_PRESSURE = 0xD0,
   MIDI_PITCH_CHANGE = 0xE0,
   MIDI_SYSEX_START  =  0xF0,
   MIDI_TIME_CODE    = 0xF1,
   MIDI_SONG_POSITION =0xF2,
   MIDI_SONG_SELECT  = 0xF3,
   MIDI_TUNE_REQUEST  = 0xF6,
   MIDI_SYSEX_STOP   = 0xF7,
   MIDI_TIMING_CLOCK = 0xF8,
   MIDI_RT_TICK = 0xF9,
   MIDI_RT_START = 0xFA,
   MIDI_RT_CONTINUE = 0xFB,
   MIDI_RT_STOP   = 0xFC,
   MIDI_RT_ACTIVE_SENSE = 0xFE,
   MIDI_RT_RESET = 0xFF
} MIDI_Commands;

#define MIDI_COMMAND_COUNT (20)

typedef enum {
   MN_C = 0,
   MN_CS,
   MN_D,
   MN_DS,
   MN_E,
   MN_F,
   MN_FS,
   MN_G,
   MN_GS,
   MN_A,
   MN_AS,
   MN_B,
   NOTE_COUNT
} MIDI_KEYS;

typedef enum {
   NO_ITEM = 0,
   ACOUSTIC_BASS_DRUM = 35,
   BASS_DRUM1,
   SIDE_STICK,
   ACOUSTIC_SNARE,
   HAND_CLAP,
   ELECTRIC_SNARE,//40
   LOW_FLOOR_TOM, 
   CLOSED_HI_HAT,
   HIGH_FLOOR_TOM,
   PEDAL_HIGH_HAT,
   LOW_TOM,
   OPEN_HIGH_HAT,
   LOW_MID_TOM,
   HI_MID_TOM,
   CRASH_CYMBAL1,
   HIGH_TOM,
   RIDE_CYMBAL1,
   CHINESE_CYMBAL,
   RIDE_BELL,
   TAMBOURINE,
   SPLASH_CYMBAL,
   COWBELL,
   CRASH_CYMBAL2,
   VIBRASLAP,
   RIDE_CYMBAL2,
   HI_BONGO,
   LOW_BONGO,
   MUTE_HI_CONGA,
   OPEN_HI_CONGA,
   LOW_CONGA,
   HIGH_TIMBALE,
   LOW_TIMBALE,
   HIGH_AGOGO,
   LOW_AGOGO,
   CABASA,
   MARACAS,
   SHORT_WHISTLE,
   LONG_WHISTLE,
   SHORT_GUIRO,
   LONG_GUIRO,
   CLAVES,
   HI_WOOD_BLOCK,
   LOW_WOOD_BLOCK,
   MUTE_CUICA,
   OPEN_CUICA,
   MUTE_TRIANGLE,
   OPEN_TRIANGLE // 83
} GENERAL_MIDI_PERCUSSION;

typedef enum {
	BANK_SELECT	=	0,
	MODULATION_WHEEL	=	1,
	BREATH_CONTROLLER	=	2,
	UNDEFINED3	=	3,
	FOOT_CONTROLLER	=	4,
	PORTAMENTO_TIME	=	5,
	DATA_ENTRY_MSB	=	6,
	CHANNEL_VOLUME	=	7,
	BALANCE	=	8,
	UNDEFINED9	=	9,
	PAN_POSITION	=	10,
	EXPRESSION_CONTROLLER	=	11,
	EFFECT_CONTROL_1	=	12,
	EFFECT_CONTROL_2	=	13,
	UNDEFINED14	=	14,
	UNDEFINED15	=	15,
	GENERAL_PURPOSE_CONTROLLER_1	=	16,
	GENERAL_PURPOSE_CONTROLLER_2	=	17,
	GENERAL_PURPOSE_CONTROLLER_3	=	18,
	GENERAL_PURPOSE_CONTROLLER_4	=	19,
	UNDEFINED20	=	20,
	UNDEFINED21	=	21,
	UNDEFINED22	=	22,
	UNDEFINED23	=	23,
	UNDEFINED24	=	24,
	UNDEFINED25	=	25,
	UNDEFINED26	=	26,
	UNDEFINED27	=	27,
	UNDEFINED28	=	28,
	UNDEFINED29	=	29,
	UNDEFINED30	=	30,
	UNDEFINED31	=	31,
	BANK_SELECT_LSB	=	32,
	MODULATION_WHEEL_LSB	=	33,
	BREATH_CONTROLLER_LSB	=	34,
	UNDEFINED35	=	35,
	FOOT_CONTROLLER_LSB	=	36,
	PORTAMENTO_TIME_LSB	=	37,
	DATA_ENTRY_LSB	=	38,
	CHANNEL_VOLUME_LSB	=	39,
	BALANCE_LSB	=	40,
	UNDEFINED41	=	41,
	PAN_POSITION_LSB	=	42,
	EXPRESSION_CONTROLLER_LSB	=	43,
	EFFECT_CONTROL_1_LSB	=	44,
	EFFECT_CONTROL_2_LSB	=	45,
	UNDEFINED46	=	46,
	UNDEFINED47	=	47,
	GENERAL_PURPOSE_CONTROLLER_1_LSB	=	48,
	GENERAL_PURPOSE_CONTROLLER_2_LSB	=	49,
	GENERAL_PURPOSE_CONTROLLER_3_LSB	=	50,
	GENERAL_PURPOSE_CONTROLLER_4_LSB	=	51,
	UNDEFINED52	=	52,
	UNDEFINED53	=	53,
	UNDEFINED54	=	54,
	UNDEFINED55	=	55,
	UNDEFINED56	=	56,
	UNDEFINED57	=	57,
	UNDEFINED58	=	58,
	UNDEFINED59	=	59,
	UNDEFINED60	=	60,
	UNDEFINED61	=	61,
	UNDEFINED62	=	62,
	UNDEFINED63	=	63,
	SUSTAIN	=	64,
	PORTAMENTO	=	65,
	SOSTENUTO	=	66,
	SOFT_PEDAL	=	67,
	LEGATO_FOOTSWITCH	=	68,
	HOLD_2	=	69,
	SOUND_VARIATION	=	70,
	RESONANCE	=	71,
	SOUND_RELEASE_TIME	=	72,
	SOUND_ATTACK_TIME	=	73,
	FREQUENCY_CUTOFF	=	74,
	DECAY_TIME	=	75,
	VIBRATO_RATE	=	76,
	VIBRATO_DEPTH	=	77,
	VIBRATO_DELAY	=	78,
	UNDEFINED79	=	79,
	GENERAL_PURPOSE_CONTROLLER_5	=	80,
	GENERAL_PURPOSE_CONTROLLER_6	=	81,
	GENERAL_PURPOSE_CONTROLLER_7	=	82,
	GENERAL_PURPOSE_CONTROLLER_8	=	83,
	PORTAMENTO_CONTROL	=	84,
	UNDEFINED85	=	85,
	UNDEFINED86	=	86,
	UNDEFINED87	=	87,
	HIGH_RESOLUTION_VELOCITY_PREFIX	=	88,
	UNDEFINED89	=	89,
	UNDEFINED90	=	90,
	REVERB_LEVEL	=	91,
	TREMOLO_LEVEL	=	92,
	CHORUS_DEPTH	=	93,
	DETUNE_DEPTH	=	94,
	PHASER_LEVEL	=	95,
	DATA_INCREMENT	=	96,
	DATA_DECREMENT	=	97,
	NON_REGISTERED_PARAMETER_LSB	=	98,
	NON_REGISTERED_PARAMETER_MSB	=	99,
	REGISTERED_PARAMETER_LSB	=	100,
	REGISTERED_PARAMETER_MSB	=	101,
	UNDEFINED102	=	102,
	UNDEFINED103	=	103,
	UNDEFINED104	=	104,
	UNDEFINED105	=	105,
	UNDEFINED106	=	106,
	UNDEFINED107	=	107,
	UNDEFINED108	=	108,
	UNDEFINED109	=	109,
	UNDEFINED110	=	110,
	UNDEFINED111	=	111,
	UNDEFINED112	=	112,
	UNDEFINED113	=	113,
	UNDEFINED114	=	114,
	UNDEFINED115	=	115,
	UNDEFINED116	=	116,
	UNDEFINED117	=	117,
	UNDEFINED118	=	118,
	UNDEFINED119	=	119,
	ALL_SOUND_OFF	=	120,
	RESET_ALL_CONTROLLERS	=	121,
	LOCAL_CONTROL	=	122,
	ALL_NOTES_OFF	=	123,
	OMNI_OFF	=	124,
	OMNI_ON	=	125,
	MONO_ON	=	126,
	POLY_ON	=	127,

} CONTROL_CHANGE_CODES;


typedef enum
{
	ACC_GRAND_PIANO	=	0	,
	BRIGHT_ACC_PIANO	=	1	,
	ELEC_GRAND_PIANO	=	2	,
	HONKY_TONK_PIANO	=	3	,
	ELECTRIC_PIANO_1	=	4	,
	ELECTRIC_PIANO_2	=	5	,
	HARPSICHORD	=	6	,
	CLAVI	=	7	,
	CELESTA	=	8	,
	GLOCKENSPIEL	=	9	,
	MUSIC_BOX	=	10	,
	VIBRAPHONE	=	11	,
	MARIMBA	=	12	,
	XYLOPHONE	=	13	,
	TUBULAR_BELLS	=	14	,
	DULCIMER	=	15	,
	DRAWBAR_ORGAN	=	16	,
	PERCUSSIVE_ORGAN	=	17	,
	ROCK_ORGAN	=	18	,
	CHURCH_ORGAN	=	19	,
	REED_ORGAN	=	20	,
	ACCORDION	=	21	,
	HARMONICA	=	22	,
	TANGO_ACCORDION	=	23	,
	NYLON_GUITAR	=	24	,
	STEEL_GUITAR	=	25	,
	JAZZ_GUITAR	=	26	,
	CLEAN_GUITAR	=	27	,
	MUTED_GUITAR	=	28	,
	OVERDRIVEN_GUITAR	=	29	,
	DISTORTION_GUITAR	=	30	,
	GUITAR_HARMONICS	=	31	,
	ACOUSTIC_BASS	=	32	,
	FINGERED_BASS	=	33	,
	PICKED_BASS	=	34	,
	FRETLESS_BASS	=	35	,
	SLAP_BASS_1	=	36	,
	SLAP_BASS_2	=	37	,
	SYNTH_BASS_1	=	38	,
	SYNTH_BASS_2	=	39	,
	VIOLIN	=	40	,
	VIOLA	=	41	,
	CELLO	=	42	,
	CONTRABASS	=	43	,
	TREMOLO_STRINGS	=	44	,
	PIZZICATO_STRINGS	=	45	,
	ORCHESTRAL_HARP	=	46	,
	TIMPANI	=	47	,
	STRING_ENSEMBLE_1	=	48	,
	STRING_ENSEMBLE_2	=	49	,
	SYNTHSTRINGS_1	=	50	,
	SYNTHSTRINGS_2	=	51	,
	CHOIR_AAHS	=	52	,
	VOICE_OOHS	=	53	,
	SYNTH_VOICE	=	54	,
	ORCHESTRA_HIT	=	55	,
	TRUMPET	=	56	,
	TROMBONE	=	57	,
	TUBA	=	58	,
	MUTED_TRUMPET	=	59	,
	FRENCH_HORN	=	60	,
	BRASS_SECTION	=	61	,
	SYNTHBRASS_1	=	62	,
	SYNTHBRASS_2	=	63	,
	SOPRANO_SAX	=	64	,
	ALTO_SAX	=	65	,
	TENOR_SAX	=	66	,
	BARITONE_SAX	=	67	,
	OBOE	=	68	,
	ENGLISH_HORN	=	69	,
	BASSOON	=	70	,
	CLARINET	=	71	,
	PICCOLO	=	72	,
	FLUTE	=	73	,
	RECORDER	=	74	,
	PAN_FLUTE	=	75	,
	BLOWN_BOTTLE	=	76	,
	SHAKUHACHI	=	77	,
	WHISTLE	=	78	,
	OCARINA	=	79	,
	SQUARE_WAVE	=	80	,
	SAW_WAVE	=	81	,
	CALLIOPE	=	82	,
	CHIFF	=	83	,
	CHARANG	=	84	,
	VOICE	=	85	,
	FIFTH_WAVE	=	86	,
	BASS_LEAD	=	87	,
	NEW_AGE	=	88	,
	WARM	=	89	,
	POLYSYNTH	=	90	,
	CHOIR	=	91	,
	BOWED	=	92	,
	METALLIC	=	93	,
	HALO	=	94	,
	SWEEP	=	95	,
	RAIN	=	96	,
	SOUNDTRACK	=	97	,
	CRYSTAL	=	98	,
	ATMOSPHERE	=	99	,
	BRIGHTNESS	=	100	,
	GOBLINS	=	101	,
	ECHOES	=	102	,
	SCI_FI	=	103	,
	SITAR	=	104	,
	BANJO	=	105	,
	SHAMISEN	=	106	,
	KOTO	=	107	,
	KALIMBA	=	108	,
	BAG_PIPE	=	109	,
	FIDDLE	=	110	,
	SHANAI	=	111	,
	TINKLE_BELL	=	112	,
	AGOGO	=	113	,
	STEEL_DRUMS	=	114	,
	WOODBLOCK	=	115	,
	TAIKO_DRUM	=	116	,
	MELODIC_TOM	=	117	,
	SYNTH_DRUM	=	118	,
	REVERSE_CYMBAL	=	119	,
	GUITAR_FRET_NOISE	=	120	,
	BREATH_NOISE	=	121	,
	SEASHORE	=	122	,
	BIRD_TWEET	=	123	,
	TELEPHONE_RING	=	124	,
	HELICOPTER	=	125	,
	APPLAUSE	=	126	,
	GUNSHOT	=	127	,

} MIDI_GM_INSTRUMENTS;


#endif
